/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: list.h,v 1.13 2005-11-07 23:22:54 zer0 Exp $
 *
 */

/** \file modules_avr/list/list.h
 *  \brief Interface of the list module
 *
 *  \todo See the todo list of list.c
 *
 *  \test See the test list of list.c
 *
 * This module provides LISTs. Don't use list size > 127. In future,
 * provide a config file for defining large lists ?
 */

#ifndef _LIST_H_
#define _LIST_H_

#define DEBUG 0
#include <stdio.h>

#define WOVERWRAPPED -1

#ifdef HOST_VERSION
#define CR "\n"
#else
#define CR "\r\n"
#endif

#include <utils.h>

/**
 * This structure is the header of a list type.
 */
struct list_hdr {
  u08 size; /**< The size of the list (number of elements) */
  u08 cur_size; /**< number of data in the list */
  u08 beg_indice; /**< indice of the first elt */
  s08 read_cursor; /**< read cursor */
} __attribute__ ((packed));

/**
 * This is a generic kind of list, in which we suppose that elements
 * are char 
*/
struct generic_list {
	struct list_hdr hdr; 
	char elt[0]; 
} __attribute__ ((packed));


/**
 * Define a new list type
 */
#define LIST_TYPEDEF(typename, elttype, size) \
typedef struct typename { \
	struct list_hdr hdr; \
	elttype elt[size]; \
} typename;

#define LIST_INIT(list, beginning) \
do { \
  list.hdr.size = sizeof(list.elt)/sizeof(list.elt[0]); \
  list.hdr.cur_size = 0; \
  list.hdr.beg_indice = beginning; \
  list.hdr.read_cursor = beginning; \
} while(0)


/**
 * Return 1 if the list is full
 */
#define LIST_FULL(list) (list.hdr.size == list.hdr.cur_size)

/**
 * Return 1 if the list is empty
 */
#define LIST_EMPTY(list) (list.hdr.cur_size == 0)

/**
 * return current size of the list (number of used elements)
 */
#define LIST_CURSIZE(list) (list.hdr.cur_size)

/**
 * return size of the list (used + free elements)
 */
#define LIST_SIZE(list) (list.hdr.size)

/**
 * return the number of free elts 
 */
#define LIST_FREESIZE(list) (list.hdr.size-list.hdr.cur_size)



#define LIST_READ_START(list, elt_p) ({ \
u08 ret=0; \
 list.hdr.read_cursor = 0 ; \
 if(elt_p) { \
    *elt_p = list.elt[list.hdr.beg_indice] ;  \
 } \
 if(DEBUG) \
   printf("LIST_READ_START(%s, %s) -> ret %d"CR,#list, #elt_p, ret); \
 ret; \
}) 

#define LIST_READ_END(list, elt_p) ({ \
u08 ret=0; \
 list.hdr.read_cursor = list.hdr.cur_size-1; \
 if(elt_p) { \
     *elt_p = list.elt[(list.hdr.beg_indice-1+list.hdr.cur_size) % list.hdr.size] ;  \
 } \
 if(DEBUG) \
   printf("LIST_READ_END(%s, %s) -> ret %d"CR,#list, #elt_p, ret); \
 ret; \
}) 


#define LIST_READ_GOTO(list, elt_p, i) ({ \
u08 ret=0; \
 if( (i<0) || (i>=list.hdr.cur_size) ) \
	 ret = EINVAL; \
 else { \
	 list.hdr.read_cursor = i; \
      if(elt_p) { \
	    *elt_p = list.elt[(list.hdr.beg_indice+i) % list.hdr.size] ;  \
      } \
 } \
 if(DEBUG) \
   printf("LIST_READ_GOTO(%s, %s, %d) -> ret %d"CR,#list, #elt_p, i, ret); \
 ret; \
}) 

#define LIST_READ_MOVE(list, elt_p, i)  ({\
u08 ret=0;  \
 if (i<0) { \
	if( (-i) > list.hdr.read_cursor ) \
		 ret = WOVERWRAPPED ; \
	list.hdr.read_cursor -= ((-i) % list.hdr.cur_size) ; \
	if (list.hdr.read_cursor < 0)  \
		list.hdr.read_cursor += list.hdr.cur_size ;  \
 } \
 else {  \
	 if( i >= list.hdr.cur_size - list.hdr.read_cursor )  \
		 ret = WOVERWRAPPED ; \
	 list.hdr.read_cursor += (i % list.hdr.cur_size) ; \
	 if (list.hdr.read_cursor >= list.hdr.cur_size) \
		 list.hdr.read_cursor -= list.hdr.cur_size ; \
 } \
 if(DEBUG) \
   printf("LIST_READ_MOVE(%s, %s, %d) -> ret %d"CR,#list, #elt_p, i, ret); \
 if(elt_p) \
	 *elt_p = list.elt[(list.hdr.beg_indice+list.hdr.read_cursor) % list.hdr.size] ;  \
 ret;  \
}) 

#define LIST_READ(list, elt_p)  ({\
u08 ret=0;  \
 if(elt_p) \
	 *elt_p = list.elt[(list.hdr.beg_indice+list.hdr.read_cursor) % list.hdr.size] ;  \
 else \
      ret=EINVAL; \
 ret;  \
}) 

#define LIST_PUSH_START(list, e) ({ \
u08 ret=0; \
 if( LIST_FULL(list) ) \
	 ret=EINVAL; \
 else { \
	 list.hdr.beg_indice = (list.hdr.beg_indice-1+list.hdr.size) % list.hdr.size; \
	 list.elt [ list.hdr.beg_indice ] = e ; \
	 list.hdr.cur_size ++ ; \
 } \
if(DEBUG) \
 printf("LIST_PUSH_START(%s, %s) -> ret %d"CR,#list, #e, ret); \
 ret; \
})

#define LIST_PUSH_END(list, e) ({ \
u08 ret=0; \
 if( LIST_FULL(list) ) \
	 ret=EINVAL; \
 else { \
	 list.elt [ (list.hdr.beg_indice+list.hdr.cur_size) % list.hdr.size ] = e ; \
	 list.hdr.cur_size ++ ; \
 } \
if(DEBUG) \
 printf("LIST_PUSH_END(%s, %s) -> ret %d"CR,#list, #e, ret); \
 ret; \
})

#define LIST_PULL_START(list, elt_p) ({ \
u08 ret=0; \
 if( LIST_EMPTY(list) ) \
	 ret=EINVAL; \
 else if (elt_p) { \
	 *elt_p = list.elt [ list.hdr.beg_indice ] ; \
	 list.hdr.beg_indice = (list.hdr.beg_indice+1) % list.hdr.size; \
	 list.hdr.cur_size -- ; \
 } \
if(DEBUG) \
 printf("LIST_PULL_START(%s, %s) -> ret %d"CR,#list, #elt_p, ret); \
 ret; \
})

#define LIST_PULL_END(list, elt_p) ({ \
u08 ret=0; \
 if( LIST_EMPTY(list) ) \
	 ret=EINVAL; \
 else if (elt_p) { \
	 *elt_p = list.elt [ (list.hdr.beg_indice-1+list.hdr.cur_size) % list.hdr.size ] ; \
	 list.hdr.cur_size -- ; \
 } \
if(DEBUG) \
 printf("LIST_PULL_END(%s, %s) -> ret %d"CR,#list, #elt_p, ret); \
 ret; \
})

/* start by the last elt */
#define LIST_ARRAY_PUSH_START(list, array, nb) ({\
 u08 ret=0; \
 s08 i; \
 if(!array) \
   ret=EINVAL; \
 else { \
   for(i=nb-1 ; (i>=0) && (!ret) ; i--) { \
      ret=LIST_PUSH_START(list, array[i]); \
   } \
 } \
 if(DEBUG) \
   printf("LIST_ARRAY_PUSH_START(%s, %s, %d) -> ret %d"CR,#list, #array, nb, ret); \
 ret; \
})

#define LIST_ARRAY_PUSH_END(list, array, nb) ({\
 u08 ret=0, i; \
 if(!array) \
   ret=EINVAL; \
 else { \
   for(i=0 ; (i<nb) && (!ret) ; i++) { \
      ret=LIST_PUSH_END(list, array[i]); \
   } \
 } \
 if(DEBUG) \
   printf("LIST_ARRAY_PUSH_END(%s, %s, %d) -> ret %d"CR,#list, #array, nb, ret); \
 ret; \
})

#define LIST_ARRAY_PULL_START(list, array, nb) ({\
 u08 ret=0, i; \
 if(!array) \
   ret=EINVAL; \
 else { \
   for(i=0 ; (i<nb) && (!ret) ; i++) { \
      ret=LIST_PULL_START(list, (array+i)); \
   } \
 } \
 if(DEBUG) \
   printf("LIST_ARRAY_PULL_START(%s, %s, %d) -> ret %d"CR,#list, #array, nb, ret); \
 ret; \
})

#define LIST_ARRAY_PULL_END(list, array, nb) ({\
 u08 ret=0; \
 s08 i; \
 if(!array) \
   ret=EINVAL; \
 else { \
   for(i=nb-1 ; (i>=0) && (!ret) ; i--) { \
      ret=LIST_PULL_END(list, (array+i)); \
   } \
 } \
 if(DEBUG) \
   printf("LIST_ARRAY_PULL_END(%s, %s, %d) -> ret %d"CR,#list, #array, nb, ret); \
 ret; \
})

#define LIST_ALIGN_LEFT(list) ({ \
u08 ret=0,i; \
if(list.hdr.beg_indice != 0) { \
	if(list.hdr.beg_indice+list.hdr.cur_size <= list.hdr.size) { \
		for(i=0 ; i<list.hdr.cur_size ; i++) { \
			list.elt[i] = list.elt[i+list.hdr.beg_indice]; \
		} \
	} \
	else { \
          u08 buffer_size=(list.hdr.size - list.hdr.beg_indice < (list.hdr.cur_size + list.hdr.beg_indice)%list.hdr.size) ? \
                           (list.hdr.size - list.hdr.beg_indice) * sizeof(list.elt[0]) : \
	                      ((list.hdr.cur_size + list.hdr.beg_indice)%list.hdr.size) * sizeof(list.elt[0]); \
		     { \
				u08 buffer[buffer_size]; \
				memcpy(buffer, list.elt, buffer_size); \
				for(i=0 ; i<(list.hdr.cur_size - buffer_size/sizeof(list.elt[0])) ; i++) { \
					list.elt[i] = list.elt[i+list.hdr.beg_indice]; \
				} \
				memcpy(&list.elt[list.hdr.cur_size - buffer_size/sizeof(list.elt[0])], buffer, buffer_size); \
		} \
	} \
 list.hdr.beg_indice=0; \
} \
 if(DEBUG) \
   printf("LIST_ALIGN_LEFT()"CR); \
 ret; \
})
#endif
