/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: f32.c,v 1.2 2006-02-05 23:58:06 zer0 Exp $
 *
 */

#include <stdio.h>
#include <f32.h>

/* value of 2^^16 in float */
#define POW2_16F (65536.0)


/**************** double <-> f32 */

f32 f32_from_double(double d)
{
    f32 f;
    f.integer = (d >= 0 ? (s16)d : (s16)d-1) ;
    f.decimal = (u16) ((d - (s16)d) * POW2_16F);
    
    return f;
}

double f32_to_double(f32 f)
{
    double d;
    
    d = f.integer;
    d += ((double)f.decimal / POW2_16F);
  
    return d;
}


/**************** s16,u16 <-> f32 */

f32 f32_from_integer(s16 i, u16 d)
{
    f32 f;
    f.integer = i;
    f.decimal = d;
    
    return f;
}

/**************** operations */

f32 f32_neg(f32 f)
{
    return s32_to_f32( -f32_to_s32(f) );
}

f32 f32_add(f32 a, f32 b)
{
    return s32_to_f32(  f32_to_s32(a) + f32_to_s32(b) );
}

f32 f32_sub(f32 a, f32 b)
{
    return s32_to_f32(  f32_to_s32(a) - f32_to_s32(b) );
}

f32 f32_inv(f32 f)
{
    if (F32_IS_ZERO(f))
        return F32_NAN;

    return s32_to_f32( ((s32)0x7fffffff) / (f32_to_s32(f)/2) );
}

f32 f32_mul(f32 a, f32 b)
{
    return s32_to_f32(   ( ((s32)(a.integer) * (s32)(b.integer)) << 16 )    +
                            (s32)(a.integer) * (s32)(b.decimal)             +
                            (s32)(a.decimal) * (s32)(b.integer)             +
                         ( ((s32)(a.decimal) * (s32)(b.decimal)) >> 16 ) );
}

f32 f32_div(f32 a, f32 b)
{
    if (F32_IS_ZERO(b))
        return F32_NAN;
    return f32_mul(a,f32_inv(b));
}

/* could be optimized... */
static inline u32  u32_sqrt(u32 x)
{
    u32 root=0x8000;
    u32 mask=0x8000;

    while(mask) {
        if( root*root < x ) {
            mask >>= 1;
            root |= mask;
        }
        else if ( root*root > x ) {
            root &= ~(mask);
            mask >>= 1;
            root |= mask;
            continue;
        }
        else 
            break;
    }
    return root;
}

f32 f32_sqrt(f32 f)
{
    u32 a,b,c,d;

    if (F32_IS_ZERO(f))
        return F32_ZERO;

    if (F32_IS_NEG(f))
        return F32_NAN;

    if(f.integer) {
        /* sqrt(a+b) = sqrt(a)*sqrt(1+b/a) */
        a=(u32)(f.integer) << 16 ;
        b=(u32)(f.decimal) << 16 ;
        c=u32_sqrt(a);
        d=u32_sqrt(0x10000 + (b/a));
        return f32_mul(s32_to_f32( c<<8 ), s32_to_f32( d<<8 ));
    }
    else {
        b=(u32)(f.decimal) << 16 ;
        return s32_to_f32(u32_sqrt(b));
    }
}

/**************** debug */

void print_f32(f32 f)
{
    if(F32_IS_EQ(f, F32_NAN))
        printf("nan");
    else
        printf("%f",f32_to_double(f));
}

