/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: f64.c,v 1.3 2006-02-21 23:05:18 zer0 Exp $
 *
 */

#include <stdio.h>
#include <f64.h>

/* value of 2^^32 in float */
#define POW2_32F (4294967296.0)


/**************** double <-> f64 */

f64 f64_from_double(double d)
{
    f64 f;
    f.integer = (d >= 0 ? (s32)d : (s32)d-1) ;
    f.decimal = (u32) ((d - (s32)d) * POW2_32F);
    
    return f;
}

double f64_to_double(f64 f)
{
    double d;
    
    d = f.integer;
    d += ((double)f.decimal / POW2_32F);
  
    return d;
}


/**************** s32,u32 <-> f64 */

f64 f64_from_integer(s32 i, u32 d)
{
    f64 f;
    f.integer = i;
    f.decimal = d;
    
    return f;
}

/**************** operations */

f64 f64_neg(f64 f)
{
    return s64_to_f64( -f64_to_s64(f) );
}

f64 f64_add(f64 a, f64 b)
{
    return s64_to_f64(  f64_to_s64(a) + f64_to_s64(b) );
}

f64 f64_sub(f64 a, f64 b)
{
    return s64_to_f64(  f64_to_s64(a) - f64_to_s64(b) );
}

f64 f64_inv(f64 f)
{
    if (F64_IS_ZERO(f))
        return F64_NAN;

    return s64_to_f64( ((s64)0x7fffffffffffffff) / (f64_to_s64(f)/2) );
}

/* about 300 us with a 16 Mhz quartz on an atm128 */
f64 f64_mul(f64 a, f64 b)
{
    return s64_to_f64(   ( ((s64)(a.integer) * (s64)(b.integer)) << 32 )    +
                            (s64)(a.integer) * (s64)(b.decimal)             +
                            (s64)(a.decimal) * (s64)(b.integer)             +
                         ( ((s64)(a.decimal) * (s64)(b.decimal)) >> 32 ) );
}

/* about 80 us with a 16 Mhz quartz on an atm128 */
s32 f64_msb_mul(f64 a, f64 b)
{
	return (s32)(( f64_to_s64(a) * f64_to_s64(b)) >> 32);
}


f64 f64_div(f64 a, f64 b)
{
    if (F64_IS_ZERO(b))
        return F64_NAN;
    return f64_mul(a,f64_inv(b));
}

/* could be optimized... */
static inline u64  u64_sqrt(u64 x)
{
    u64 root=0x80000000;
    u64 mask=0x80000000;

    while(mask) {
        if( root*root < x ) {
            mask >>= 1;
            root |= mask;
        }
        else if ( root*root > x ) {
            root &= ~(mask);
            mask >>= 1;
            root |= mask;
            continue;
        }
        else 
            break;
    }
    return root;
}

f64 f64_sqrt(f64 f)
{
    u64 a,b,c,d;

    if (F64_IS_ZERO(f))
        return F64_ZERO;

    if (F64_IS_NEG(f))
        return F64_NAN;

    if(f.integer) {
        /* sqrt(a+b) = sqrt(a)*sqrt(1+b/a) */
        a=(u64)(f.integer) << 32 ;
        b=(u64)(f.decimal) << 32 ;
        c=u64_sqrt(a);
        d=u64_sqrt(0x100000000 + (b/a));
        return f64_mul(s64_to_f64( c<<16 ), s64_to_f64( d<<16 ));
    }
    else {
        b=(u64)(f.decimal) << 32 ;
        return s64_to_f64(u64_sqrt(b));
    }
}

/**************** debug */

void print_f64(f64 f)
{
    if(F64_IS_EQ(f, F64_NAN))
        printf("nan");
    else
        printf("%f",f64_to_double(f));
}

