/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: scheduler.h,v 1.7 2005-10-29 17:30:36 zer0 Exp $
 *
 */

/* Eirbot & Microb Technology 2005 - Zer0
 * Interface of the SCHEDULER Module
 */

/** \file scheduler.h
 *  \brief Interface of the SCHEDULER Module
 *
 *  \todo multi arch ?
 *
 *  \test nothing done
 *
 * This module provides a function scheduler. You can call
 * scheduler_add_event for adding a function to the scheduler, and
 * specifying what interval between each call. During the execution of
 * the function, interrupts are masked !! So use this module with
 * caution (small functions) for avoiding problems.
 *
 * Functions with a high priority value will be call before others
 * (default is 128).
 *
 * This module uses Timer 0
 */


#ifndef _SCHEDULER_H_
#define _SCHEDULER_H_

#include <utils.h>
#include <scheduler_config.h>

#define SCHEDULER_MCU_CLOCK (F_CPU / KHz)

/** TIME_UNIT is the number of microseconds between each interruption
 * if the prescaler equals 1 */
#define TIMER_UNIT ( 256000l / (SCHEDULER_MCU_CLOCK) )

/** SCHEDULER_UNIT is the REAL number of microseconds between each
 *  interruption. We can use it like this :
 *    scheduler_add_periodical_event(f, 1000l/SCHEDULER_UNIT);
 *  The function f will be called every ms. 
 */
#define SCHEDULER_UNIT ( TIMER_UNIT * SCHEDULER_CLOCK_PRESCALER )

#define SCHEDULER_PERIODICAL 0
#define SCHEDULER_SINGLE 1

#define SCHEDULER_DEFAULT_PRIORITY 128

#define scheduler_add_single_event_priority(f, data, period, prio) \
    __scheduler_add_event(SCHEDULER_SINGLE, f, data, period, prio);

#define scheduler_add_periodical_event_priority(f, data, period, prio) \
    __scheduler_add_event(SCHEDULER_PERIODICAL, f, data, period, prio);

#define scheduler_add_single_event(f, data, period) \
    __scheduler_add_event(SCHEDULER_SINGLE, f, data, period, SCHEDULER_DEFAULT_PRIORITY);

#define scheduler_add_periodical_event(f, data, period) \
    __scheduler_add_event(SCHEDULER_PERIODICAL, f, data, period, SCHEDULER_DEFAULT_PRIORITY);

/** Initialisation of the module */
void scheduler_init(void);

/** 
 * Adds an event to the event table.
 * Returns the id of the event on succes and -1 on error
 * You can use macro above for simpler use.
 */
s08 __scheduler_add_event(u08 unicity, void (*f)(void *), void * data, u16 period, u08 priority);

/**
 * Dels an event from the table by its ID. If there is no event,
 * nothing is done.
 */
s08 scheduler_del_event(s08 num);

#endif
