/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: irq_lock_macros.h,v 1.6 2006-02-05 23:58:06 zer0 Exp $
 *
 */

/** \file modules/base/utils/irq_lock_macros.h
 *  \brief Interface of the utils module
 *   
 *   here are defined the three macros : 
 *
 *  IRQ_LOCK();              this locks the interrupts
 *  IRQ_UNLOCK();            this unlocks the interrupts
 *  IRQ_UNLOCK_STACKED();    this is to be used if the unlock is not at the same identation level
 *  
 *  code example follows: 
 *  
 *    IRQ_LOCK();
 *      // code to be protected against interrupts ...
 *    IRQ_UNLOCK(); // needs to be associated with an unlock
 *  
 *  example of the use of the three macros : 
 *  
 *    IRQ_LOCK();
 *      if(I_want_to_unlock) {
 *        IRQ_UNLOCK_STACKED(); // not the same indentation level
 *  return;
 *      }
 *    // code some times protected against interrupts ...
 *    IRQ_UNLOCK();
 *
 *
 * Note that these macros exists as functions (for MICROB one) :
 * irq_lock() and irq_unlock() (in lowercase)
 */


#ifdef HOST_VERSION

#define IRQ_UNLOCK() 
#define IRQ_LOCK() 

void irq_lock(void);
void irq_unlock(void);

#else

#define GLOBAL_IRQ_ARE_MASKED() (!(bit_is_set(SREG,7)))


#ifdef USE_MICROB_IRQ /* first version */

/** 
 * intr_cpt is the interrupt counter used to count the number of irq
 * locks (ONLY IN CASE OF USE_MICROB_IRQ is defined). If it is 0, no
 * lock has been put. BUT : irq can be masked or not by cli() (global
 * irq flag).
 * 
 * The absolute vaule of the counter stores the number of locks.
 * 
 * The sign determines wether the first lock was done with global
 * interrupt disable or enable. WARNING : if you use IRQ_LOCK/UNLOCK
 * you should not use sei() and cli() (only one sei() call at the
 * beginning of your code)
 *
 * Example :
 * 
 * sei()           -> intr_cpt = 0
 * IRQ_LOCK()      -> intr_cpt = 1
 * IRQ_LOCK()      -> intr_cpt = 2
 * IRQ_UNLOCK()    -> intr_cpt = 1
 * IRQ_UNLOCK()    -> intr_cpt = 0
 * 
 * cli()           -> intr_cpt = 0
 * IRQ_LOCK()      -> intr_cpt = -1
 * IRQ_LOCK()      -> intr_cpt = -2
 * IRQ_UNLOCK()    -> intr_cpt = -1
 * IRQ_LOCK()      -> intr_cpt = -2
 * IRQ_UNLOCK()    -> intr_cpt = -1
 * IRQ_UNLOCK()    -> intr_cpt = 0
 *
 */
extern volatile u08 intr_cpt;




#define IRQ_NB_LOCK() (intr_cpt)


/* if intrps are set
 *     cli
 *     if intr_cpt < 0 (-> not normal)
 *         intr_cpt --
 *     else
 *         intr_cpt ++
 * else (intrps are not set)
 *     if intr_cpt <= 0 
 *         intr_cpt --
 *     else 
 *         intr_cpt ++
 */
#define IRQ_LOCK() asm volatile(                                                                                      \
                     "brbc 7, l_lock_b%="           "\n\t" /* if SREG[7]==0, jump to l_lock_b */                      \
                     "cli"                          "\n\t" /* (INTERRUPTS WERE ON) else do mask intrp */              \
                     "ld __tmp_reg__, %a0"          "\n\t" /* load intrp_cpt in a temp register (r0 normally) */      \
                     "cp __tmp_reg__, __zero_reg__" "\n\t" /* compare intrp_cpt with 0 */                             \
                     "brlt l_lock_a%="              "\n\t" /* if intrp_cpt < 0 jump to l_lock_a */                    \
                     "inc __tmp_reg__"              "\n\t" /* else intrp_cpt > 0, increment intrp_cpt */              \
                     "rjmp l_lock_d%="              "\n"   /* jump to l_lock_d (end) */                               \
"l_lock_a%=:" "\n\t" "dec __tmp_reg__"              "\n\t" /* decrement intrp_cpt */                                  \
                     "rjmp l_lock_d%="              "\n"   /* jump to l_lock_d (end) */                               \
"l_lock_b%=:" "\n\t" "ld __tmp_reg__, %a0"          "\n\t" /* (INTERRUPTS WERE OFF) load intrp_cpt in a temp reg */   \
                     "cp __zero_reg__, __tmp_reg__" "\n\t" /* compare intrp_cpt with 0 */                             \
                     "brge l_lock_c%="              "\n\t" /* if intrp_cpt <= 0 jump to l_lock_c */                   \
                     "inc __tmp_reg__"              "\n\t" /* else intrp_cpt > 0, increment intrp_cpt */              \
                     "rjmp l_lock_d%="              "\n"   /* jump to l_lock_d */                                     \
"l_lock_c%=:" "\n\t" "dec __tmp_reg__"              "\n"   /* decrement intrp_cpt */                                  \
"l_lock_d%=:" "\n\t" "st %a0, __tmp_reg__"          "\n\t" /* save intrp_cpt value */                                 \
                    :                                                                                                 \
                    : "e" ((volatile u08 *)&intr_cpt) )



/* if intrps are set    (-> not normal)
 *     cli
 *     if intr_cpt < 0
 *         intr_cpt++
 *     if intr_cpt > 0
 *         intr_cpt--
 *     sei
 * else (intrps are not set)
 *     if intr_cpt < 0
 *         intr_cpt++
 *     if intr_cpt > 0
 *         intr_cpt--
 *         if intr_cpt == 0
 *             sei
 */
#define IRQ_UNLOCK() asm volatile(                                                                                    \
                     "brbc 7, l_ulck_c%="           "\n\t"   /* if SREG[7]==0, jump to l_ulck_c */                    \
                     "cli"                          "\n\t" /* (INTERRUPTS WERE ON) else do mask intrp */              \
                     "ld __tmp_reg__, %a0"          "\n\t" /* load intrp_cpt in a temp register (r0 normally) */      \
                     "cp __tmp_reg__, __zero_reg__" "\n\t" /* compare intrp_cpt with 0 */                             \
                     "brlt l_ulck_b%="              "\n\t" /* if intrp_cpt < 0 jump to l_ulck_b */                    \
                     "breq l_ulck_a%="              "\n\t" /* if intrp_cpt == 0 jump to l_ulck_a */                   \
                     "dec __tmp_reg__"              "\n"   /* else intrp_cpt > 0, decrement intrp_cpt */              \
"l_ulck_a%=:" "\n\t" "sei"                          "\n\t" /* sei */                                                  \
                     "rjmp l_ulck_d%="              "\n"   /* jump to l_ulck_d (end) */                               \
"l_ulck_b%=:" "\n\t" "inc __tmp_reg__"              "\n\t" /* increment intrp_cpt */                                  \
                     "sei"                          "\n\t" /* sei */                                                  \
                     "rjmp l_ulck_e%="              "\n"   /* jump to l_ulck_e (end) */                               \
"l_ulck_c%=:" "\n\t" "ld __tmp_reg__, %a0"          "\n\t" /* (INTERRUPTS WERE OFF) load intrp_cpt in a temp reg */   \
                     "cp __tmp_reg__, __zero_reg__" "\n\t" /* compare intrp_cpt with 0 */                             \
                     "brlt l_ulck_d%="              "\n\t" /* if intrp_cpt < 0 jump to l_ulck_d */                    \
                     "breq l_ulck_e%="              "\n\t" /* if intrp_cpt == 0 jump to l_ulck_e */                   \
                     "dec __tmp_reg__"              "\n\t" /* else intrp_cpt > 0, decrement intrp_cpt */              \
                     "brne l_ulck_e%="              "\n\t" /* if intrp_cpt != 0 jump to l_ulck_e */                   \
                     "st %a0, __tmp_reg__"          "\n\t" /* save intrp_cpt value */                                 \
                     "sei"                          "\n\t" /* sei */                                                  \
                     "rjmp l_ulck_f%="              "\n"   /* jump to l_ulck_f */                                     \
"l_ulck_d%=:" "\n\t" "inc __tmp_reg__"              "\n"   /* increment intrp_cpt */                                  \
"l_ulck_e%=:" "\n\t" "st %a0, __tmp_reg__"          "\n\t" /* save intrp_cpt value */                                 \
"l_ulck_f%=:" "\n\t"                                       /* end */                                                  \
                    :                                                                                                 \
                    : "e" ((volatile u08 *)&intr_cpt) )

/* for compatibility */
#define IRQ_UNLOCK_STACKED() IRQ_UNLOCK()

void irq_lock(void);
void irq_unlock(void);

#else /* second version... */

#define IRQ_LOCK() do {              \
  u08 interrup_status = SREG;        \
  cli();


#define IRQ_UNLOCK() SREG = interrup_status;  \
  } while ( 0 )

#define IRQ_UNLOCK_STACKED() SREG = interrup_status;

#define irq_lock()   IRQ_LOCK()
#define irq_unlock() IRQ_UNLOCK()

#endif /* ! USE_MICROB_IRQ */

#endif /* ! HOST_VERSION */
