/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: utils.h,v 1.33 2006-02-05 23:58:06 zer0 Exp $
 *
 */

/** \file modules/base/utils/utils.h
 *  \brief Interface of the utils module
 *
 * here are some cute little macros, and other stuff, microcontroller
 * related ! 
 */


#ifndef _UTILS_H_
#define _UTILS_H_

#ifndef HOST_VERSION
#include <avr/io.h>
#endif

#define F_CPU ((unsigned long)CONFIG_QUARTZ)

#include <utils_config.h>
#include <hertz.h>
#include <utils_types.h>
#include <utils_errno.h>

#include <autoconf.h>
#include <irq_lock_macros.h>


/*
 *  a few "mathematical" macros : maximums and minimums
 */

/**
 *  signed maxmimum : both signs are tested
 */
#define S_MAX(to_saturate, value_max)    \
do {                                     \
   if (to_saturate > value_max)          \
     to_saturate = value_max;            \
   else if (to_saturate < -value_max)    \
     to_saturate = -value_max;           \
 } while(0)

/**
 *  unsigned maximum : result >0 is forced
 */
#define U_MAX(to_saturate, value_max)    \
do {                                     \
   if (to_saturate > value_max)          \
     to_saturate = value_max;            \
   else if (to_saturate < 0)             \
     to_saturate = 0;                    \
 } while(0)

/**
 *   simple maximum
 */
#define MAX(to_saturate, value_max)      \
do {                                     \
   if (to_saturate > value_max)          \
     to_saturate = value_max;            \
} while(0)

/**
 *  simple minimum
 */
#define MIN(to_saturate, value_min)      \
do {                                     \
   if (to_saturate < value_min)          \
     to_saturate = value_min;            \
} while(0)


/** absolute
 *  while the abs() function in the libc works only with int type
 *  this macro works with every numerical type including floats
 */
#define ABS( val) ( ((val) <0) ? -(val) : (val) )


/* byte extraction, not recommended for use
 *  use only if you need speed optimization !
 *  use ">>" instead for current operations 
 */
#define extr16_08_0(i) (*(char *)(&i))         // LSB of a 16bit
#define extr16_08_1(i) (*((char *)(&i)+1))     // MSB of a 16bit

#define extr32_16_0(i) (*(int *)(&i))          // LSB of a 32 bit
#define extr32_16_1(i) (* (((int *)(&i)) +1))  // MSB of a 32 bit
#define extr32_16_23(i) (*((int *)((char *)(&i)+1))) // middle of a 32 bit

#define extr32_08_0(i) (*(char *)(&i))         // same stuff
#define extr32_08_1(i) (*((char *)(&i)+1))
#define extr32_08_2(i) (*((char *)(&i)+2))
#define extr32_08_3(i) (*((char *)(&i)+3))


/* a few asm utilities */
#define nop() __asm__ __volatile__ ("NOP\n") ; /** nop instruction, 1 CPU cycle consumed */
#define rien() __asm__ __volatile__ (" \n") ;  /** nothing */
#define cli() __asm__ __volatile__ ("CLI\n") ; /** disable interrupts */
#define sei() __asm__ __volatile__ ("SEI\n") ; /** enable interrupts */

#define reset() __asm__ __volatile__ ("jmp 0\n") ; /** simple software reset, but doesn't initialize the registers */

/**
 *   little bit toggeling macro 
 *  
 *  change pin state
 *  usage :
 *  BIT_TOGGLE(PORTB,2) to make the pin 2 of PORTB toggle
 */
#define BIT_TOGGLE(port,bit) do {\
      if(bit_is_set(PIN(port),bit)) \
	cbi(port,bit); \
      else \
	sbi(port,bit); \
      } while(0)


/** macros eeprom */
#ifndef HOST_VERSION
#include <avr/eeprom.h>

#define eeprom_r08(addr) eeprom_read_byte((u08 *)(addr))

#define eeprom_r16(addr) eeprom_read_word((uint16_t *)(addr))

extern u32 eeprom_r32(u32 * addr);
extern u32 eeprom_r24(u32 * addr);
#endif // not HOST_VERSION


/** booleans */
#define FALSE 0
#define TRUE 1
#define False FALSE
#define false FALSE
#define True TRUE
#define true TRUE


/** DDR and PINS from port adress */
#define DDR(port) (*(&(port) -1))
#define PIN(port) (*(&(port) -2))

/** open collector simulation macros */
#define OPEN_CO_INIT(port, bit) sbi(port,bit)
#define OPEN_CO_HIGH(port, bit) cbi(DDR(port),bit)
#define OPEN_CO_LOW(port, bit)  cbi(DDR(port),bit)

/** deprecated macros in libc, but they're almost used, so we implement them again ;) */
#ifndef cbi
#define cbi(sfr, bit) ( sfr &= ~ _BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) ( sfr |= _BV(bit))
#endif


#endif /* ifndef _UTILS_H_ */

