/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: uart.c,v 1.29 2006-02-21 23:05:18 zer0 Exp $
 *
 */

/* Olivier MATZ, Droids-corp 2004 - 2006
 * Implementation of the uart module
 */

/** \file uart.c
 *  \brief Implementation of the UART module.
 *
 *  \todo check uart9bits without intr, check parity, raise errors if necessary,
 *        no fifo, tx events are not always called, need to add some config
 */

#include <avr/io.h>
#include <avr/signal.h>

#include <utils.h>
#include <list.h>

#include <uart_config.h>
#include <uart_defs.h>
#include <uart_errors.h>
#include <uart.h>


typedef void (event)(char);
typedef void (event_9bits)(int);

/******************************************
 * UART variables
 */

#define DEFINE_VARS(x)                                        \
                                                              \
/** The emission fifo of uart ##x## */                        \
LIST_TYPEDEF(tx##x##_fifo_t, char, UART##x##_TX_FIFO_SIZE);   \
tx##x##_fifo_t g_tx##x##_fifo;                                \
                                                              \
/** The reception fifo of uart ##x## */                       \
LIST_TYPEDEF(rx##x##_fifo_t, char, UART##x##_RX_FIFO_SIZE);   \
rx##x##_fifo_t g_rx##x##_fifo;                                \
                                                              \
event * rx##x##_event = NULL;                                 \
event * tx##x##_event = NULL;

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_VARS(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_VARS(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_VARS(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_VARS(3)
#endif


/******************************************
 * UART private function declarations
 */

#define DECLARE_FCTS(x)                                        \
static inline u08 uart##x##_getconf_nbits(void);


#if (defined UDR0) && (defined UART0_COMPILE)
DECLARE_FCTS(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DECLARE_FCTS(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DECLARE_FCTS(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DECLARE_FCTS(3)
#endif



#if (defined UDR0) && (defined UART0_COMPILE)
/**
 * This is the interruption function which occurs when the entire
 * frame in the transmit shift register has been shifted out and 
 * there is no new data in the transmit buffer.
 */
SIGNAL(SIG_UART0_DATA);

/**
 * This is the interruption function which occurs when there is
 * a new unread data in the reception buffer.
 */
SIGNAL(SIG_UART0_RECV);
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
SIGNAL(SIG_UART1_DATA);
SIGNAL(SIG_UART1_RECV);
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
SIGNAL(SIG_UART2_DATA);
SIGNAL(SIG_UART2_RECV);
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
SIGNAL(SIG_UART3_DATA);
SIGNAL(SIG_UART3_RECV);
#endif


/******************************************
 * UART UDR access
 */

#define DEFINE_8BITS_UDR_FCTS(x)                    \
static inline char uart##x##_get_udr(void)          \
{                                                   \
  return UDR##x;                                    \
}                                                   \
                                                    \
static inline void uart##x##_set_udr(char c)        \
{                                                   \
  UDR##x = c;                                       \
}                                                   \

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_8BITS_UDR_FCTS(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_8BITS_UDR_FCTS(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_8BITS_UDR_FCTS(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_8BITS_UDR_FCTS(3)
#endif



#define DEFINE_9BITS_UDR_FCTS(x)                    \
                                                    \
static inline int uart##x##_get_udr_9bits(void)     \
{                                                   \
  int val = UDR##x;                                 \
  val |= ( UCSR##x##B & ( 1 << RXB8) ? 0x100 : 0 ); \
  return val;                                       \
}                                                   \
                                                    \
static inline void uart##x##_set_udr_9bits(int c)   \
{                                                   \
  if( c & 0x100 )                                   \
    sbi(UCSR##x##B,RXB8);                           \
  else                                              \
    cbi(UCSR##x##B,RXB8);                           \
  UDR##x = c;                                       \
}

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_9BITS_UDR_FCTS(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_9BITS_UDR_FCTS(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_9BITS_UDR_FCTS(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_9BITS_UDR_FCTS(3)
#endif


/******************************************
 * UART TX Interrupt
 */

#define DEFINE_TRANS_INTRPT(x)                                     \
                                                                   \
/** emit next char of fifo if any, and call the event function *   \
 * this function is executed with intr locked                  */  \
static inline void uart##x##_send_next_char(void)                  \
{                                                                  \
                                                                   \
  if (uart##x##_getconf_nbits() == 9) {                            \
    int elt = 0;                                                   \
                                                                   \
    /* for 9 bits, it uses 2 places in the fifo */                 \
    if(LIST_CURSIZE(g_tx##x##_fifo) < 2) {                         \
      cbi(UCSR##x##B, UDRIE);                                      \
      return;                                                      \
    }                                                              \
                                                                   \
    LIST_PULL_END(g_tx##x##_fifo, &extr16_08_0(elt));              \
    LIST_PULL_END(g_tx##x##_fifo, &extr16_08_1(elt));              \
                                                                   \
    uart##x##_set_udr_9bits(elt);                                  \
    sbi(UCSR##x##B, UDRIE);                                        \
                                                                   \
    /* tx event function. We suppose interrupts are already */     \
    /* locked, so no pb with tx##x##_event pointer */              \
    if(tx##x##_event)                                              \
      ((event_9bits *)tx##x##_event)(elt);                         \
                                                                   \
  }                                                                \
  else { /* 5, 6, 7 or 8 bits */                                   \
    char elt = 0;                                                  \
                                                                   \
    if(LIST_EMPTY(g_tx##x##_fifo)) {                               \
      cbi(UCSR##x##B, UDRIE);                                      \
      return;                                                      \
    }                                                              \
                                                                   \
    LIST_PULL_END(g_tx##x##_fifo, &elt);                           \
    uart##x##_set_udr(elt);                                        \
    sbi(UCSR##x##B, UDRIE);                                        \
                                                                   \
    /* tx event function. We suppose interrupts are already */     \
    /* locked, so no pb with tx##x##_event pointer */              \
    if(tx##x##_event)                                              \
      tx##x##_event(elt);                                          \
  }                                                                \
}                                                                  \
                                                                   \
/* if the transmission of the byte is finished, send next char */  \
/* of fifo */                                                      \
SIGNAL(SIG_UART##x##_DATA)                                         \
{                                                                  \
  uart##x##_send_next_char();                                      \
}

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_TRANS_INTRPT(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_TRANS_INTRPT(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_TRANS_INTRPT(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_TRANS_INTRPT(3)
#endif


/******************************************
 * UART RX Interrupt
 */

#define DEFINE_RECV_INTRPT(x)                                      \
                                                                   \
SIGNAL(SIG_UART##x##_RECV)                                         \
{                                                                  \
  if (uart##x##_getconf_nbits() == 9) {                            \
    int elt = 0;                                                   \
                                                                   \
    elt = uart##x##_get_udr_9bits();                               \
    if ( LIST_FREESIZE(g_rx##x##_fifo) >= 2 ) {                    \
      LIST_PUSH_START(g_rx##x##_fifo, extr16_08_0(elt)); /*lsb*/   \
      LIST_PUSH_START(g_rx##x##_fifo, extr16_08_1(elt)); /*msb*/   \
    }                                                              \
                                                                   \
    /* rx event function. We suppose interrupts are already */     \
    /* locked, so no pb with rx##x##_event pointer */              \
    if(rx##x##_event)                                              \
      ((event_9bits *)rx##x##_event)(elt);                         \
  }                                                                \
  else {                                                           \
    char elt = 0;                                                  \
                                                                   \
    elt = uart##x##_get_udr();                                     \
    if ( ! LIST_FULL(g_rx##x##_fifo) ) {                           \
      LIST_PUSH_START(g_rx##x##_fifo, elt);                        \
    }                                                              \
                                                                   \
    /* rx event function. We suppose interrupts are already */     \
    /* locked, so no pb with rx##x##_event pointer */              \
    if(rx##x##_event)                                              \
      rx##x##_event(elt);                                          \
  }                                                                \
}

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_RECV_INTRPT(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_RECV_INTRPT(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_RECV_INTRPT(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_RECV_INTRPT(3)
#endif



/******************************************
 * UART USER SEND/RECV
 */

#define DEFINE_IO_FCTS(x)                                                   \
                                                                            \
/* send a char, or put it in the fifo if uart is not ready.  */             \
/* Return -1 if fifo is full */                                             \
int uart##x##_send_nowait(char c)                                           \
{                                                                           \
                                                                            \
  irq_lock();                                                               \
                                                                            \
  /* if uart intrp mode is disabled (note that we look rx */                \
  /* intrp -- RXCIE is 0) */                                                \
  if ( ! ( UCSR##x##B & ( 1 << RXCIE ) ) ) {                                \
      /* we have to poll the status register before xmit */                 \
      if( UCSR##x##A & (1<<UDRE) ) {                                        \
        uart##x##_set_udr(c);                                               \
        irq_unlock();                                                       \
        return (int)c;                                                      \
      }                                                                     \
      else {                                                                \
        irq_unlock();                                                       \
        return -1;                                                          \
      }                                                                     \
  }                                                                         \
                                                                            \
  if( LIST_FULL(g_tx##x##_fifo) ) {                                         \
    /* exit */                                                              \
    irq_unlock();                                                           \
    return -1;                                                              \
  }                                                                         \
                                                                            \
  /* uart is ready to send */                                               \
  if( LIST_EMPTY(g_tx##x##_fifo) && UCSR##x##A & (1<<UDRE) ) {              \
    uart##x##_set_udr(c);                                                   \
    sbi( UCSR##x##B, UDRIE ) ;                                              \
  }                                                                         \
  else { /* not ready, put char in fifo */                                  \
    LIST_PUSH_START(g_tx##x##_fifo, c);                                     \
  }                                                                         \
                                                                            \
  irq_unlock();                                                             \
  return (int)c;                                                            \
}                                                                           \
                                                                            \
int uart##x##_send(char c)                                                  \
{                                                                           \
  /* if cannot send the char */                                             \
  if (uart##x##_send_nowait(c) == -1) {                                     \
                                                                            \
  /* if irq lock are masked, we have to poll the status */                  \
  /* register */                                                            \
    if( GLOBAL_IRQ_ARE_MASKED() ) {                                         \
      while( !(UCSR##x##A & (1<<UDRE)) );                                   \
      /* send the next char in the fifo to free a place */                  \
      uart##x##_send_next_char();                                           \
      LIST_PUSH_START(g_tx##x##_fifo, c);                                   \
    }                                                                       \
    else {                                                                  \
      /* if irq are not locked, we can loop to emit */                      \
      while(uart##x##_send_nowait(c) == -1);                                \
    }                                                                       \
  }                                                                         \
  return c;                                                                 \
}                                                                           \
                                                                            \
int uart##x##_send_9bits_nowait(int c)                                      \
{                                                                           \
                                                                            \
  irq_lock();                                                               \
                                                                            \
  /* if tx intrp are disabled (RXCIE is 0) */                               \
  if ( ! ( UCSR##x##B & ( 1 << RXCIE ) ) ) {                                \
      /* we have to poll the status register before xmit */                 \
      if( (UCSR##x##A & (1<<UDRE)) ) {                                      \
        uart##x##_set_udr_9bits(c);                                         \
        irq_unlock();                                                       \
        return c;                                                           \
      }                                                                     \
      else {                                                                \
        irq_unlock();                                                       \
        return -1;                                                          \
      }                                                                     \
  }                                                                         \
                                                                            \
  /* the fifo must have 2 free places */                                    \
  if( LIST_FREESIZE(g_tx##x##_fifo) < 2) {                                  \
    /* exit */                                                              \
    irq_unlock();                                                           \
    return -1;                                                              \
  }                                                                         \
                                                                            \
  /* uart is ready to send */                                               \
  if( LIST_EMPTY(g_tx##x##_fifo) && UCSR##x##A & (1<<UDRE) ) {              \
    uart##x##_set_udr_9bits(c);                                             \
    sbi( UCSR##x##B, UDRIE ) ;                                              \
  }                                                                         \
  else { /* not ready, put char in fifo */                                  \
    LIST_PUSH_START(g_tx##x##_fifo, extr16_08_0(c));                        \
    LIST_PUSH_START(g_tx##x##_fifo, extr16_08_1(c));                        \
  }                                                                         \
                                                                            \
  irq_unlock();                                                             \
  return (int)c;                                                            \
}                                                                           \
                                                                            \
int uart##x##_send_9bits(int c)                                             \
{                                                                           \
  /* if cannot send the char */                                             \
  if (uart##x##_send_9bits_nowait(c) == -1) {                               \
                                                                            \
  /* if irq lock are masked, we have to poll the status */                  \
  /* register */                                                            \
    if( GLOBAL_IRQ_ARE_MASKED() ) {                                         \
      while( !(UCSR##x##A & (1<<UDRE)) );                                   \
      /* send the next char in the fifo to free two places */               \
      uart##x##_send_next_char();                                           \
      LIST_PUSH_START(g_tx##x##_fifo, extr16_08_0(c));                      \
      LIST_PUSH_START(g_tx##x##_fifo, extr16_08_1(c));                      \
    }                                                                       \
    else {                                                                  \
      /* if irq are not locked, we can loop to emit */                      \
      while(uart##x##_send_9bits_nowait(c) == -1);                          \
    }                                                                       \
  }                                                                         \
  return c;                                                                 \
}                                                                           \
                                                                            \
/* get a char from the receive fifo */                                      \
int uart##x##_recv_nowait(void)                                             \
{                                                                           \
  char elt;                                                                 \
                                                                            \
  irq_lock();                                                               \
  if( ! LIST_EMPTY(g_rx##x##_fifo) ) {                                      \
    LIST_PULL_END(g_rx##x##_fifo, &elt);                                    \
    irq_unlock();                                                           \
    return (u16)elt;                                                        \
  }                                                                         \
  irq_unlock();                                                             \
  return (u16)(-1);                                                         \
}                                                                           \
                                                                            \
/* get a char from the receive fifo */                                      \
int uart##x##_recv(void)                                                    \
{                                                                           \
  int elt;                                                                  \
  while ( (elt = uart##x##_recv_nowait()) == -1 );                          \
  return elt;                                                               \
}                                                                           \
                                                                            \
/* get a char from the receive fifo */                                      \
int uart##x##_9bits_recv_nowait(void)                                       \
{                                                                           \
  char elt;                                                                 \
                                                                            \
  irq_lock();                                                               \
  if( LIST_CURSIZE(g_rx##x##_fifo) >= 2) {                                  \
    LIST_PULL_END(g_rx##x##_fifo, &extr16_08_0(elt));                       \
    LIST_PULL_END(g_rx##x##_fifo, &extr16_08_1(elt));                       \
    irq_unlock();                                                           \
    return (u16)elt;                                                        \
  }                                                                         \
  irq_unlock();                                                             \
  return (u16)(-1);                                                         \
}                                                                           \
                                                                            \
/* get a char from the receive fifo */                                      \
int uart##x##_9bits_recv(void)                                              \
{                                                                           \
  int elt;                                                                  \
  while ( (elt = uart##x##_9bits_recv_nowait()) == -1 );                    \
  return elt;                                                               \
}                                                                           \
                                                                            
#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_IO_FCTS(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_IO_FCTS(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_IO_FCTS(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_IO_FCTS(3)
#endif


/******************************************
 * UART set conf
 */

#if UART_IS_USART
#define DEFINE_SET_NBITS_PARITY(x)                                  \
                                                                    \
static inline s08 uart##x##_set_nbits_parity(struct uart_config * u)\
{                                                                   \
    u08 ucsrc##x = 0;                                               \
                                                                    \
    /* number of bit in the frame */                                \
    if((u->nbits < 5) || (u->nbits > 9)) {                          \
      return ENOTSUP;                                               \
    }                                                               \
    ucsrc##x |= ( ((u->nbits - 5) & 0x03) << UCSZ0 ) ;              \
    if ( u->nbits == 9 )                                            \
      UCSR##x##B |= ( 1 << UCSZ2 );                                 \
    else                                                            \
      UCSR##x##B &= ( 0xFE << UCSZ2 );                              \
                                                                    \
    /* parity */                                                    \
    if(u->parity == UART_PARTITY_ODD)                               \
      ucsrc##x |= ( ( 1 << UPM0 ) | ( 1 << UPM1 ) ) ;               \
    else if(u->parity == UART_PARTITY_EVEN)                         \
      ucsrc##x |= ( 1 << UPM1 ) ;                                   \
    else if(u->parity != UART_PARTITY_NONE) {                       \
        return EINVAL;                                              \
    }                                                               \
                                                                    \
    /* nb of stop bits */                                           \
    if(u->stop_bits == UART_STOP_BITS_2)                            \
        ucsrc##x |= ( 1 << USBS ) ;                                 \
    else if (u->stop_bits != UART_STOP_BITS_1)  {                   \
        return EINVAL;                                              \
    }                                                               \
                                                                    \
    /* here we use a macro because some uC use a special bit */     \
    /* URSEL to access to UCSRC */                                  \
    SET_UCSR##x##C(ucsrc##x);                                       \
                                                                    \
    return ESUCCESS;                                                \
}

#else
#define DEFINE_SET_NBITS_PARITY(x)                                  \
                                                                    \
static inline s08 uart##x##_set_nbits_parity(struct uart_config * u)\
{                                                                   \
    /* number of bit in the frame */                                \
    if(u->nbits == 8)                                               \
      UCSR##x##B &= (u08)(0xFE << CHR9);                            \
    else {                                                          \
      UCSR##x##B |= (u08)(1 << CHR9);                               \
      return ENOTSUP;                                               \
    }                                                               \
                                                                    \
    /* parity and stop */                                           \
    if( (u->parity != UART_PARTITY_NONE) ||                         \
	(u->stop_bits != UART_STOP_BITS_1) ) {                      \
      return ENOTSUP;                                               \
    }                                                               \
                                                                    \
    return ESUCCESS;                                                \
}
#endif

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_SET_NBITS_PARITY(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_SET_NBITS_PARITY(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_SET_NBITS_PARITY(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_SET_NBITS_PARITY(3)
#endif


#if UART_IS_USART
#define DEFINE_SET_BAUDREG(x)                                      \
                                                                   \
static inline s08 uart##x##_set_baudreg(u16 baudreg)               \
{                                                                  \
  u08 lo,hi;                                                       \
                                                                   \
  /* set msb bit of hi to 0 (useful fot uC with URSEL, and */      \
  /* not important for the others because the baudreg */           \
  /* will never be as big */                                       \
  lo=(u08)baudreg;                                                 \
  hi=(u08)(baudreg>>8) & 0x7F;                                     \
                                                                   \
  UBRR##x##H=hi;                                                   \
  UBRR##x##L=lo;                                                   \
                                                                   \
  return ESUCCESS;                                                 \
}

#else

#define DEFINE_SET_BAUDREG(x)                                      \
                                                                   \
static inline s08 uart##x##_set_baudreg(u16 baudreg)               \
{                                                                  \
  u08 lo ,hi;                                                      \
                                                                   \
  lo=(u08)baudreg;                                                 \
  hi=(u08)(baudreg>>8);                                            \
                                                                   \
  if(hi != 0) {                                                    \
      return EINVAL;                                               \
  }                                                                \
  UBRR##x##L=lo;                                                   \
                                                                   \
  return ESUCCESS;                                                 \
}
#endif


#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_SET_BAUDREG(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_SET_BAUDREG(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_SET_BAUDREG(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_SET_BAUDREG(3)
#endif


#define DEFINE_SET_CONF(x)                                      \
                                                                \
s08 uart##x##_setconf(struct uart_config * u)                   \
{                                                               \
  u08 ret = ESUCCESS;                                           \
  u16 baudrate_reg;                                             \
  struct uart_config static_conf;                               \
                                                                \
  irq_lock();                                                   \
                                                                \
  /* static configuration */                                    \
  if (!u) {                                                     \
    u = &static_conf;                                           \
    u->enabled = UART##x##_ENABLED ;                            \
    u->intr_enabled = UART##x##_INTERRUPT_ENABLED ;             \
    u->use_double_speed = UART##x##_USE_DOUBLE_SPEED ;          \
    u->parity =  UART##x##_PARITY ;                             \
    u->stop_bits = UART##x##_STOP_BIT ;                         \
    u->nbits = UART##x##_NBITS ;                                \
    u->baudrate = UART##x##_BAUDRATE ;                          \
  }                                                             \
                                                                \
  /* wait xmit finished (UDRE = 1) */                           \
  while(! (UCSR##x##A & (1<<UDRE)) );                           \
                                                                \
  LIST_INIT(g_tx##x##_fifo, 0);                                 \
  LIST_INIT(g_rx##x##_fifo, 0);                                 \
                                                                \
  UCSR##x##A=0;                                                 \
                                                                \
  if(u->enabled)                                                \
    UCSR##x##B = (( 1 << TXEN  ) | ( 1 << RXEN ));              \
  else {                                                        \
    UCSR##x##B = 0;                                             \
    goto out; /* no more conf */                                \
  }                                                             \
                                                                \
  /* we only enable recv interrupt, the xmit intrpt will be */  \
  /* enabled in the xmit function */                            \
  if(u->intr_enabled)                                           \
    UCSR##x##B |= ( 1 << RXCIE );                               \
                                                                \
  if(UART_HAS_U2X) { /* if u2x is supported */                  \
    if(u->use_double_speed) /* u2x is enabled */                \
      UCSR##x##A |= ( 1 << U2X );                               \
    else                                                        \
      UCSR##x##A &= ( 0xFE << U2X );                            \
  }                                                             \
  else if(u->use_double_speed) {                                \
    ret = ENOTSUP;                                              \
    goto out;                                                   \
  }                                                             \
                                                                \
  uart##x##_set_nbits_parity(u);                                \
                                                                \
  /* baudrate */                                                \
  if(u->use_double_speed)                                       \
      baudrate_reg = (F_CPU / (u->baudrate*8l))-1;              \
  else                                                          \
      baudrate_reg = (F_CPU / (u->baudrate*16l))-1;             \
                                                                \
  uart##x##_set_baudreg(baudrate_reg);                          \
                                                                \
  /* exit */                                                    \
 out:                                                           \
  irq_unlock();                                                 \
  return ret;                                                   \
}

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_SET_CONF(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_SET_CONF(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_SET_CONF(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_SET_CONF(3)
#endif



/******************************************
 * UART get conf
 */

#if UART_IS_USART
#define DEFINE_GET_NBITS(x)                                     \
                                                                \
/* return number of bits in current conf */                     \
static inline u08 uart##x##_getconf_nbits(void)                 \
{                                                               \
  u08 nbits;                                                    \
                                                                \
  nbits = (GET_UCSR##x##C() >> UCSZ0) & 0x03 ;                  \
  if ( UCSR##x##B & ( 1 << UCSZ2 ) )                            \
      nbits += 4;                                               \
  nbits += 5 ;                                                  \
                                                                \
  return nbits;                                                 \
}

#else

#define DEFINE_GET_NBITS(x)                                     \
                                                                \
/* return number of bits in current conf */                     \
static inline u08 uart##x##_getconf_nbits(void)                 \
{                                                               \
  u08 nbits;                                                    \
                                                                \
  if ( UCSR##x##B & (u08)(1 << CHR9) )                          \
      return 8;                                                 \
  else                                                          \
      return 9;                                                 \
}
#endif

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_GET_NBITS(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_GET_NBITS(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_GET_NBITS(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_GET_NBITS(3)
#endif


#if UART_IS_USART
#define DEFINE_GET_BAUDREG(x)                                    \
                                                                 \
/* return number of bits in current conf */                      \
static inline u16 uart##x##_get_baudreg(void)                    \
{                                                                \
  return ((u16)UBRR##x##H)<<8 | (u16)UBRR##x##L;                 \
}

#else

#define DEFINE_GET_BAUDREG(x)                                    \
                                                                 \
/* return number of bits in current conf */                      \
static inline u16 uart##x##_get_baudreg(void)                    \
{                                                                \
  return UBRR##x##L;                                             \
}
#endif

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_GET_BAUDREG(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_GET_BAUDREG(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_GET_BAUDREG(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_GET_BAUDREG(3)
#endif



#define DEFINE_GET_CONF(x)                                                     \
                                                                               \
/* get the running uart configurtion */                                        \
void uart##x##_getconf(struct uart_config * u)                                 \
{                                                                              \
  u08 tmp;                                                                     \
                                                                               \
  irq_lock();                                                                  \
                                                                               \
  /* enabled if RXEN is set */                                                 \
  if ( UCSR##x##B & (1 << RXEN ))                                              \
    u->enabled = 1;                                                            \
  else                                                                         \
    u->enabled = 0;                                                            \
                                                                               \
  /* intrp enabled if RXCIE is set */                                          \
  if ( UCSR##x##B & (1 << RXCIE ))                                             \
    u->intr_enabled = 1;                                                       \
  else                                                                         \
    u->intr_enabled = 0;                                                       \
                                                                               \
  /* use double speed */                                                       \
  if(UART_HAS_U2X && (UCSR##x##A & ( 1 << U2X ) ) )                            \
    u->use_double_speed = 1;                                                   \
  else                                                                         \
    u->use_double_speed = 0;                                                   \
                                                                               \
                                                                               \
  /* parity */                                                                 \
  if (UART_IS_USART) {                                                         \
    tmp = GET_UCSR##x##C() & ( ( 1 << UPM1 ) | ( 1 << UPM0 ) ) ;               \
    if ( tmp == (( 1 << UPM1 ) | ( 1 << UPM0 )) )                              \
      u->parity = UART_PARTITY_ODD;                                            \
    else if ( tmp == ( 1 << UPM1 ) )                                           \
      u->parity = UART_PARTITY_EVEN;                                           \
    else                                                                       \
      u->parity = UART_PARTITY_NONE;                                           \
  }                                                                            \
  else {                                                                       \
    u->parity = UART_PARTITY_NONE;                                             \
  }                                                                            \
                                                                               \
  /* stop_bits */                                                              \
  if  (UART_IS_USART && (GET_UCSR##x##C() & ( 1 << USBS ) ) ) {                \
    u->stop_bits = UART_STOP_BITS_2;                                           \
  }                                                                            \
  else {                                                                       \
    u->stop_bits = UART_STOP_BITS_1;                                           \
  }                                                                            \
                                                                               \
  /* nbits */                                                                  \
  u->nbits = uart##x##_getconf_nbits();                                        \
                                                                               \
  u->baudrate = ( F_CPU / ((uart##x##_get_baudreg()+1) * 16) ) ;               \
                                                                               \
  irq_unlock();                                                                \
}

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_GET_CONF(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_GET_CONF(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_GET_CONF(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_GET_CONF(3)
#endif


/******************************************
 * UART INIT
 */

/* init all uarts */
void uart_init(void)
{
#if (defined UDR0) && (defined UART0_COMPILE)
    uart0_setconf(NULL);
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
    uart1_setconf(NULL);
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
    uart2_setconf(NULL);
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
    uart3_setconf(NULL);
#endif
}


/******************************************
 * UART  EVENTS
 */

#define DEFINE_EVENTS(x)                                                 \
                                                                         \
/* This function is used to register another function which will be */   \
/* executed at each byte transmission. */                                \
void uart##x##_register_tx_event(void (*f)(char))                        \
{                                                                        \
  irq_lock();                                                            \
  tx##x##_event = f;                                                     \
  irq_unlock();                                                          \
}                                                                        \
                                                                         \
/* This function is used to register another function which will be */   \
/* executed at each byte reception */                                    \
void uart##x##_register_rx_event(void (*f)(char))                        \
{                                                                        \
  irq_lock();                                                            \
  rx##x##_event = f;                                                     \
  irq_unlock();                                                          \
}

#if (defined UDR0) && (defined UART0_COMPILE)
DEFINE_EVENTS(0)
#endif

#if (defined UDR1) && (defined UART1_COMPILE)
DEFINE_EVENTS(1)
#endif

#if (defined UDR2) && (defined UART2_COMPILE)
DEFINE_EVENTS(2)
#endif

#if (defined UDR3) && (defined UART3_COMPILE)
DEFINE_EVENTS(3)
#endif

