/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: control_system_manager.c,v 1.4 2006-02-05 23:58:07 zer0 Exp $
 *
 */
#include <stdio.h>
#include <control_system_manager.h>

#define DEBUG 0

#if DEBUG == 1
#define debug_printf(args...) do { printf(args); } while(0)
#else
#define debug_printf(args...) do { } while(0)
#endif

/** Call a filter() pointer : 
 * - lock the interrupts
 * - read the pointer to the filter function
 * - unlock the interrupts
 * - if pointer is null, return the IN value
 * - else apply filter
 */
#define SAFE_FILTER( f, param, value ) ({     \
    s32 (*__f)(void *, s32);                  \
    s32 __tmp = value;                        \
    void * __param ;                          \
    irq_lock();                               \
    __f = f;                                  \
    __param = param;                          \
    irq_unlock();                             \
    if ( __f ) {                              \
        __tmp = __f(__param, value);          \
    }                                         \
    __tmp; \
})

/** Call a processout() pointer : 
 * - lock the interrupts
 * - read the pointer to the processout function
 * - unlock the interrupts
 * - if pointer is null, return 0
 * - else return the value processed by the function
 */
#define SAFE_GETPROCESSOUT( f, param ) ({     \
    s32 (*__f)(void *);                       \
    s32 __tmp = 0;                            \
    void * __param ;                          \
    irq_lock();                               \
    __f = f;                                  \
    __param = param;                          \
    irq_unlock();                             \
    if ( __f ) {                              \
        __tmp = __f(__param);                 \
    }                                         \
    __tmp; \
})

/** Call a processin() pointer : 
 * - lock the interrupts
 * - read the pointer to the processin function
 * - unlock the interrupts
 * - if pointer is null, don't do anything
 * - else call the processin with the parameters
 */
#define SAFE_SETPROCESSIN( f, param, value ) do {      \
    void (*__f)(void *, s32);                          \
    void * __param ;                                   \
    irq_lock();                                        \
    __f = f;                                           \
    __param = param;                                   \
    irq_unlock();                                      \
    if ( __f ) {                                       \
        __f(__param, value);                           \
    }                                                  \
} while(0)

/**********************************************/

void cs_init(struct cs* cs)
{
    irq_lock();
    cs->consign_filter = NULL;
    cs->consign_filter_params = NULL;

    cs->correct_filter = NULL;
    cs->correct_filter_params = NULL;

    cs->feedback_filter = NULL;
    cs->feedback_filter_params = NULL;

    cs->process_out = NULL;
    cs->process_out_params = NULL;

    cs->process_in = NULL;
    cs->process_in_params = NULL;

    cs->consign_value = 0;
    cs->error_value = 0;
    cs->out_value = 0;
    irq_unlock();

    return;
}


void cs_set_consign_filter(struct cs* cs, s32 (*consign_filter)(void*, s32), void* consign_filter_params)
{
    irq_lock();
    cs->consign_filter = consign_filter;
    cs->consign_filter_params = consign_filter_params;
    irq_unlock();
}



void cs_set_correct_filter(struct cs* cs, s32 (*correct_filter)(void*, s32), void* correct_filter_params)
{
    irq_lock();
    cs->correct_filter = correct_filter;
    cs->correct_filter_params = correct_filter_params;
    irq_unlock();
}


void cs_set_feedback_filter(struct cs* cs, s32 (*feedback_filter)(void*, s32), void* feedback_filter_params)
{
    irq_lock();
    cs->feedback_filter = feedback_filter;
    cs->feedback_filter_params = feedback_filter_params;
    irq_unlock();
}


void cs_set_process_in(struct cs* cs, void (*process_in)(void*, s32), void* process_in_params)
{        
    irq_lock();
    cs->process_in = process_in;
    cs->process_in_params = process_in_params;
    irq_unlock();
}



void cs_set_process_out(struct cs* cs, s32 (*process_out)(void*), void* process_out_params)
{
    irq_lock();
    cs->process_out = process_out;
    cs->process_out_params = process_out_params;
    irq_unlock();
}



s32 cs_do_process(struct cs* cs, s32 consign)
{
#if DEBUG == 1
    static int i=0;
#endif
    s32 process_out_value = 0;

    /* save the consign value into the structure */
    cs->consign_value = consign;

    debug_printf("%d %ld ", i++, consign);

    /* if the consign filter exist */
    consign = SAFE_FILTER(cs->consign_filter, cs->consign_filter_params, consign);

    debug_printf("%ld ", consign);

    /* read the process out if defined */
    process_out_value = SAFE_GETPROCESSOUT(cs->process_out, cs->process_out_params);

    debug_printf("%ld ", process_out_value);

    /* apply the feedback filter if defined */
    process_out_value = SAFE_FILTER(cs->feedback_filter, cs->feedback_filter_params, process_out_value);

    debug_printf("%ld ", process_out_value);

    /* substract consign and process out and put it into error */
    cs->error_value = consign - process_out_value ;
    
    debug_printf("%ld ", cs->error_value);

    /* apply the correct filter to error_value and put it into out_value */
    cs->out_value = SAFE_FILTER(cs->correct_filter, cs->correct_filter_params, cs->error_value);
 
    debug_printf("%ld\n", cs->out_value);

    /* send out_value to process in*/
    SAFE_SETPROCESSIN (cs->process_in, cs->process_in_params, cs->out_value);

    /* return the out value */
    return (cs->out_value);
}



void cs_manage(void * data)
{
    struct cs* cs = data;
    cs_do_process(cs, cs->consign_value);
}



s32 cs_get_out(struct cs* cs)
{
    s32 tmp;
    irq_lock();
    tmp = cs->out_value;
    irq_unlock();

    return tmp;
}



s32 cs_get_error(struct cs* cs)
{
    s32 tmp;
    irq_lock();
    tmp = cs->error_value;
    irq_unlock();

    return tmp;
}



s32 cs_get_consign(struct cs* cs)
{
    s32 tmp;
    irq_lock();
    tmp = cs->consign_value;
    irq_unlock();

    return tmp;
}



void cs_set_consign(struct cs* cs, s32 v)
{
    /* set the consign */
    irq_lock();
    cs->consign_value = v;
    irq_unlock();
}        
