/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: encoders_microb.c,v 1.3 2006-02-21 23:05:18 zer0 Exp $
 *
 */

/** \file encoders_microb.c
 *  \brief Implementation for getting motor position
 *
 *  \todo test the module
 *
 *  \test nothing done
 *
 * This module provides functions for getting position from
 * coders. Its implementation depends on hardware, so that's why there
 * is (or will be) multiple implementation of this module.
 * 
 * The harware is quite simple here :
 * 
 * \verbatim         
 * ----------
 * |        |  8
 * |    PINx|<-/---  value
 * |        |   
 * |        |  8
 * |    PINy|<-/---  value
 * |        |
 * |        |  8
 * |    PINz|<-/---  value
 * |        |
 * |    ....|
 * |        |
 * |    SEL |---->  (1 bit)
 * |        |  
 * |        |
 * |        |
 * ----------
 * \endverbatim         
 */

#include <stdlib.h>

#include <utils.h>
#include <wait.h>
#include <f64.h>

#include <encoders_microb.h>
#include <encoders_microb_config.h>


/******************************************/
s32 g_encoders_microb_values[ENCODERS_MICROB_NUMBER];
s08 g_encoders_microb_previous[ENCODERS_MICROB_NUMBER];

/******************************************/

#define SEL_A() cbi(ENCODERS_MICROB_SEL_PORT,ENCODERS_MICROB_SEL_BIT)
#define SEL_B() sbi(ENCODERS_MICROB_SEL_PORT,ENCODERS_MICROB_SEL_BIT)

/******************************************/

#define READ(n)                                     \
do {                                                \
val = ENCODERS_MICROB##n##_PIN;                     \
nop();                                              \
if (val != ENCODERS_MICROB##n##_PIN)                \
     val = ENCODERS_MICROB##n##_PIN;                \
                                                    \
res = (val - g_encoders_microb_previous[n]);        \
g_encoders_microb_previous[n] = val;                \
                                                    \
g_encoders_microb_values[n] += (s32)res ;           \
} while(0)

/******************************************/

#define INIT(n)   do {                              \
g_encoders_microb_values[n] = 0;                    \
} while(0)

/******************************************/





/** 
 * Initialisation of encoders, variables
 */
void encoders_microb_init(void)
{
  sbi(DDR(ENCODERS_MICROB_SEL_PORT),ENCODERS_MICROB_SEL_BIT);
  SEL_A();
  // wait at least 10us
  wait_4cyc(50);
  encoders_microb_manage(NULL);
  wait_4cyc(50);
  encoders_microb_manage(NULL);

  IRQ_LOCK();
#ifdef ENCODERS_MICROB0_ENABLED
  INIT(0);
#endif

#ifdef ENCODERS_MICROB2_ENABLED
  INIT(2);
#endif

#ifdef ENCODERS_MICROB4_ENABLED
  INIT(4);
#endif

#ifdef ENCODERS_MICROB6_ENABLED
  INIT(6);
#endif

  
  SEL_B();
  // wait at least 10us
  wait_4cyc(50);

#ifdef ENCODERS_MICROB1_ENABLED
  INIT(1);
#endif

#ifdef ENCODERS_MICROB3_ENABLED
  INIT(3);
#endif

#ifdef ENCODERS_MICROB5_ENABLED
  INIT(5);
#endif

#ifdef ENCODERS_MICROB7_ENABLED
  INIT(7);
#endif
  IRQ_UNLOCK();

}


/** 
 * Update encoders values, need to be done quite often
 * (Fmax_encoders/64). First, encoders 0 2 4 6 are read, and at next call
 * encoders 1 3 5 7. 
 * 
 * \param dummy : a (void *) pointer that is not used. It is here according
 *                to the encoders interface.
 */
void encoders_microb_manage(void * dummy)
{
  static u08 sel=0;

  s08 val, res;

  IRQ_LOCK();

  if ( ! sel) /* here SEL_A() is set since last call */
    {
#ifdef ENCODERS_MICROB0_ENABLED
      READ(0);
#endif
      
#ifdef ENCODERS_MICROB2_ENABLED
      READ(2);
#endif
      
#ifdef ENCODERS_MICROB4_ENABLED
      READ(4);
#endif
      
#ifdef ENCODERS_MICROB6_ENABLED
      READ(6);
#endif

      SEL_B();
    }
  else /* here SEL_B() is set since last call */
    {
#ifdef ENCODERS_MICROB1_ENABLED
      READ(1);
#endif
      
#ifdef ENCODERS_MICROB3_ENABLED
      READ(3);
#endif
      
#ifdef ENCODERS_MICROB5_ENABLED
      READ(5);
#endif
      
#ifdef ENCODERS_MICROB7_ENABLED
      READ(7);
#endif

      SEL_A();
    }
  
  sel = !sel;

  IRQ_UNLOCK();

}



/** Extract encoder value.
 *
 * \param number : a (void *) that is casted in (int) containing the number
 *                 of the encoder to be read.
 */
s32 encoders_microb_get_value(void * number)
{
  s32 value;

  IRQ_LOCK();
  value = g_encoders_microb_values[(int)number];
  IRQ_UNLOCK();

  return value;
}

/** Set an encoder value
 *
 * \param number : a (void *) that is casted in (int) containing the number
 *                 of the encoder to be read.
 * \param v      : the value
 */
void encoders_microb_set_value(void * number, s32 v)
{
  IRQ_LOCK();
  g_encoders_microb_values[(int)number] = v;
  IRQ_UNLOCK();
}


/** Extract encoder value. The value is adjusted by the gain defined in
 * the config file : encoders_microb_config.h
 *
 * \param number : a (void *) that is casted in (int) containing the number
 *                 of the encoder to be read.
 */
s32 encoders_microb_get_value_adjusted(void * number)
{
  f64 f64_value;
  s32 int_value;

  IRQ_LOCK();
  int_value = g_encoders_microb_values[(int)number];
  IRQ_UNLOCK();

  f64_value = f64_from_integer(int_value, 0);

  switch((int)number)
    {
#ifdef ENCODERS_MICROB0_ENABLED
    case 0:
      return f64_msb_mul(f64_value, (f64_from_double(ENCODERS_MICROB0_GAIN)));
#endif

#ifdef ENCODERS_MICROB1_ENABLED
    case 1:
      return f64_msb_mul(f64_value, (f64_from_double(ENCODERS_MICROB1_GAIN)));
#endif

#ifdef ENCODERS_MICROB2_ENABLED
    case 2:
      return f64_msb_mul(f64_value, (f64_from_double(ENCODERS_MICROB2_GAIN)));
#endif

#ifdef ENCODERS_MICROB3_ENABLED
    case 3:
      return f64_msb_mul(f64_value, (f64_from_double(ENCODERS_MICROB3_GAIN)));
#endif

#ifdef ENCODERS_MICROB4_ENABLED
    case 4:
      return f64_msb_mul(f64_value, (f64_from_double(ENCODERS_MICROB4_GAIN)));
#endif

#ifdef ENCODERS_MICROB5_ENABLED
    case 5:
      return f64_msb_mul(f64_value, (f64_from_double(ENCODERS_MICROB5_GAIN)));
#endif

#ifdef ENCODERS_MICROB6_ENABLED
    case 6:
      return f64_msb_mul(f64_value, (f64_from_double(ENCODERS_MICROB6_GAIN)));
#endif

#ifdef ENCODERS_MICROB7_ENABLED
    case 7:
      return f64_msb_mul(f64_value, (f64_from_double(ENCODERS_MICROB7_GAIN)));
#endif

    default:
      return 0;
    }

  /* never reached normally */
  return 0;
}

