/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: position_manager.c,v 1.3 2006-02-21 23:05:18 zer0 Exp $
 *
 */

#include <string.h>
#include <math.h>

#include <robot_system.h>
#include <position_manager.h>

/** initialization of the robot_position pos, everthing is set to 0 */
void position_init(struct robot_position * pos)
{
	irq_lock();
	memset(pos, 0, sizeof(struct robot_position));
	irq_unlock();
}

/** Set a new robot position */
void position_set(struct robot_position * pos, s16 x, s16 y, s16 a)
{
	irq_lock();
	pos->pos_d.a = a;
	pos->pos_d.x = x;
	pos->pos_d.y = y;
	pos->pos_s16.x = x;
	pos->pos_s16.y = y;
	pos->pos_s16.a = a;
	irq_unlock();
}

/** 
 * Save in pos structure the pointer to the associated robot_system. 
 * The robot_system structure is used to get values from virtual encoders
 * that return angle and distance.
 */
void position_set_related_robot_system(struct robot_position * pos, struct robot_system * rs)
{
	irq_lock();
	pos->rs = rs;
	irq_unlock();
}

/** 
 * Set the physical parameters of the robot : 
 *  - number of impulsions for 1 cm (distance)
 *  - number of impulsions for 1 degree (angle)
 */
void position_set_physical_params(struct robot_position * pos, double track_cm,
				  double distance_imp_per_cm)
{
	irq_lock();
	pos->phys.track_cm = track_cm;
	pos->phys.distance_imp_per_cm = distance_imp_per_cm;
	irq_unlock();
}

/** 
 * Process the absolute position (x,y,a) depending on the delta on
 * virtual encoders since last read, and depending on physical
 * parameters. The processed position is in cm.
 */
void position_manage(struct robot_position * pos)
{
	double x, y, a;
	s16 x_s16, y_s16, a_s16;
	struct rs_polar encoders;
	struct rs_polar delta;
	struct robot_system * rs;
 
	irq_lock();
	if( !(rs = pos->rs) ) {
		irq_unlock();
		/* here we could raise an error */
		return;
	}

	irq_unlock();
	
	encoders.distance = rs_get_distance(rs);
	encoders.angle = rs_get_angle(rs);

	/* process difference between 2 measures.
	 * No lock for prev_encoders since we are the only one to use
	 * this var */
	delta.distance = encoders.distance - pos->prev_encoders.distance;
	delta.angle = encoders.angle - pos->prev_encoders.angle;

	pos->prev_encoders = encoders;

	/* update double position */
	irq_lock();
	a = pos->pos_d.a;
	x = pos->pos_d.x;
	y = pos->pos_d.y;
	irq_unlock();
  
	a += (double) delta.angle / (pos->phys.track_cm * pos->phys.distance_imp_per_cm/2);

	if(a < -M_PI)
		a += (M_PI*2);
	else if (a > (M_PI))
		a -= (M_PI*2);

	x += cos(a) * ((double) delta.distance / (pos->phys.distance_imp_per_cm)) ;
	y += sin(a) * ((double) delta.distance / (pos->phys.distance_imp_per_cm)) ;

	/* update int position */
	x_s16 = (s16)x;
	y_s16 = (s16)y;
	a_s16 = (s16)(a * (360.0/(M_PI*2)));

	irq_lock();
	pos->pos_d.a = a;
	pos->pos_d.x = x;
	pos->pos_d.y = y;
	pos->pos_s16.x = x_s16;
	pos->pos_s16.y = y_s16;
	pos->pos_s16.a = a_s16;
	irq_unlock();
}


/**
 * returns current x
 */
inline s16 position_get_x_s16(struct robot_position *pos)
{
	return pos->pos_s16.x;
}

/**
 * returns current y
 */
inline s16 position_get_y_s16(struct robot_position *pos)
{
	return pos->pos_s16.y;
}

/**
 * returns current alpha
 */
inline s16 position_get_a_deg_s16(struct robot_position *pos)
{
	return pos->pos_s16.a;
}

/********* double */

/**
 * returns current x
 */
inline double position_get_x_double(struct robot_position *pos)
{
	return pos->pos_d.x;
}

/**
 * returns current y
 */
inline double position_get_y_double(struct robot_position *pos)
{
	return pos->pos_d.y;
}

/**
 * returns current alpha
 */
inline double position_get_a_rad_double(struct robot_position *pos)
{
	return pos->pos_d.a;
}

