/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: robot_system.c,v 1.3 2006-02-21 23:05:18 zer0 Exp $
 *
 */

/**
 * The goal of this module is to provide an interface to motor and
 * encoders of the robot. This module provide a function that returns
 * the value of virtual encoders containing distance and angle. It
 * also allow the user to command virtual angle and distance PWMs.
 */

#include <string.h>
#include <stdio.h>

#include <aversive_error.h>

#include <utils.h>
#include <f64.h>

#include "angle_distance.h"
#include "robot_system.h"


/** Call a pwm() pointer : 
 * - lock the interrupts
 * - read the pointer to the pwm function
 * - unlock the interrupts
 * - if pointer is null, don't do anything
 * - else call the pwm with the parameters
 */
#define SAFE_SETPWM( f, param, value ) do {            \
    void (*__f)(void *, s32);                          \
    void * __param ;                                   \
    irq_lock();                                        \
    __f = f;                                           \
    __param = param;                                   \
    irq_unlock();                                      \
    if ( __f ) {                                       \
        __f(__param, value);                           \
    }                                                  \
} while(0)

/** Call a encoder() pointer : 
 * - lock the interrupts
 * - read the pointer to the encoder function
 * - unlock the interrupts
 * - if pointer is null, return 0
 * - else return the value processed by the function
 */
#define SAFE_GETENCODER( f, param ) ({        \
    s32 (*__f)(void *);                       \
    s32 __tmp = 0;                            \
    void * __param ;                          \
    irq_lock();                               \
    __f = f;                                  \
    __param = param;                          \
    irq_unlock();                             \
    if ( __f ) {                              \
        __tmp = __f(__param);                 \
    }                                         \
    __tmp; \
})

/** Set the structure to 0 */
void rs_init( struct robot_system * rs)
{
	irq_lock();
	memset(rs, 0, sizeof(struct robot_system));
	rs_set_ratio(rs, 1.0);
	irq_unlock();
}

/** define ratio between mot and ext track. (track_mot / track_ext) */
void rs_set_ratio(struct robot_system * rs, double ratio)
{
	irq_lock();
	rs->ratio_mot_ext = f64_from_double(ratio);
	irq_unlock();
}

/** define left PWM function and param */
void rs_set_left_pwm(struct robot_system * rs, void (*left_pwm)(void *, s32), void *left_pwm_param)
{
	irq_lock();
	rs->left_pwm = left_pwm;
	rs->left_pwm_param = left_pwm_param;
	irq_unlock();
}

/** define right PWM function and param */
void rs_set_right_pwm(struct robot_system * rs, void (*right_pwm)(void *, s32), void *right_pwm_param)
{
	irq_lock();
	rs->right_pwm = right_pwm;
	rs->right_pwm_param = right_pwm_param;
	irq_unlock();
}

/** define left motor encoder function and param */
void rs_set_left_mot_encoder(struct robot_system * rs, s32 (*left_mot_encoder)(void *), 
			     void *left_mot_encoder_param, double gain)
{
	irq_lock();
	rs->left_mot_encoder = left_mot_encoder;
	rs->left_mot_encoder_param = left_mot_encoder_param;
	rs->left_mot_gain = f64_from_double(gain);
	irq_unlock();
}

/** define right motor encoder function and param */
void rs_set_right_mot_encoder(struct robot_system * rs, s32 (*right_mot_encoder)(void *), 
			      void *right_mot_encoder_param, double gain)
{
	irq_lock();
	rs->right_mot_encoder = right_mot_encoder;
	rs->right_mot_encoder_param = right_mot_encoder_param;
	rs->right_mot_gain = f64_from_double(gain);
	irq_unlock();
}

/** define left external encoder function and param */
void rs_set_left_ext_encoder(struct robot_system * rs, s32 (*left_ext_encoder)(void *), 
			     void *left_ext_encoder_param, double gain)
{
	irq_lock();
	rs->left_ext_encoder = left_ext_encoder;
	rs->left_ext_encoder_param = left_ext_encoder_param;
	rs->left_ext_gain = f64_from_double(gain);
	irq_unlock();
}

/** define right external encoder function and param */
void rs_set_right_ext_encoder(struct robot_system * rs, s32 (*right_ext_encoder)(void *), 
			      void *right_ext_encoder_param, double gain)
{
	irq_lock();
	rs->right_ext_encoder = right_ext_encoder;
	rs->right_ext_encoder_param = right_ext_encoder_param;
	rs->right_ext_gain = f64_from_double(gain);
	irq_unlock();
}

/**** Virtual encoders and PWM */

/** 
 * set the real pwms according to the specified angle (it also
 * depends on the last distance command sent) 
 */
void rs_set_angle(void * data, s32 angle)
{
	struct rs_polar p;
	struct rs_wheels w;
	struct robot_system * rs = data;

	irq_lock();
	p.distance = rs->virtual_pwm.distance ;
	rs->virtual_pwm.angle = angle;
	irq_unlock();

	p.angle = angle;
	w = rs_get_wheels_from_polar(p);
	
	SAFE_SETPWM(rs->left_pwm, rs->left_pwm_param, w.left);
	SAFE_SETPWM(rs->right_pwm, rs->right_pwm_param, w.right);
}

/** 
 * set the real pwms according to the specified distance (it also
 * depends on the last angle command sent) 
 */
void rs_set_distance(void * data, s32 distance)
{
	struct robot_system * rs = data;
	struct rs_polar p;
	struct rs_wheels w;

	irq_lock();
	p.angle = rs->virtual_pwm.angle ;
	rs->virtual_pwm.distance = distance;
	irq_unlock();

	p.distance = distance;
	w = rs_get_wheels_from_polar(p);
	
	SAFE_SETPWM(rs->left_pwm, rs->left_pwm_param, w.left);
	SAFE_SETPWM(rs->right_pwm, rs->right_pwm_param, w.right);
}

/** 
 * get the virtual angle according to real encoders value. 
 */
s32 rs_get_angle(void * data)
{
	struct robot_system * rs = data;
	s32 angle;
	
	irq_lock();
	angle = rs->virtual_encoders.angle ;	
	irq_unlock();
	return angle;
}

/** 
 * get the virtual distance according to real encoders value. 
 */
s32 rs_get_distance(void * data)
{
	struct robot_system * rs = data;
	s32 distance;
	
	irq_lock();
	distance = rs->virtual_encoders.distance ;	
	irq_unlock();
	return distance;
}


/** 
 * Read the encoders, and update internal virtual counters. Call this
 * function is needed before reading the virtual encoders.The program
 * will decide if it the external encoders or the motor encoders are
 * taken in account (depending on flags, but not yet)
 */
void rs_update(void * data)
{
	struct robot_system * rs = data;
	struct rs_wheels wext, wmot;
	struct rs_polar pext, pmot;
	
	/* FOR THE MOMENT, WE ONLY USE EXT ENCODERS */
	wext.left = SAFE_GETENCODER(rs->left_ext_encoder, rs->left_ext_encoder_param);
	wext.right = SAFE_GETENCODER(rs->right_ext_encoder, rs->right_ext_encoder_param);

	wmot.left = SAFE_GETENCODER(rs->left_mot_encoder, rs->left_mot_encoder_param);
	wmot.right = SAFE_GETENCODER(rs->right_mot_encoder, rs->right_mot_encoder_param);
	
	DEBUG(E_ROBOT_SYSTEM, "ENCODERS : %ld %ld %ld %ld", wmot.left, wmot.right, wext.left, wext.right);

	wext.left = f64_msb_mul(f64_from_integer(0,wext.left), rs->left_ext_gain);
	wext.right = f64_msb_mul(f64_from_integer(0,wext.right), rs->right_ext_gain);
	
	wmot.left = f64_msb_mul(f64_from_integer(0,wmot.left), rs->left_mot_gain);
	wmot.right = f64_msb_mul(f64_from_integer(0,wmot.right), rs->right_mot_gain);
	
	pext = rs_get_polar_from_wheels(wext);
	pmot = rs_get_polar_from_wheels(wmot);
	
	pext.angle = f64_msb_mul(f64_from_integer(0,pext.angle), rs->ratio_mot_ext);

	irq_lock();
	rs->virtual_encoders = pext;
	irq_unlock();
}
