/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: robot_system.h,v 1.3 2006-02-21 23:05:18 zer0 Exp $
 *
 */

/**
 * The goal of this module is to provide an interface to motor and
 * encoders of the robot. This module provide a function that returns
 * the value of virtual encoders containing distance and angle. It
 * also allow the user to command virtual angle and distance PWMs.
 */

#include <utils.h>
#include <f64.h>

#include "angle_distance.h"

#ifndef _ROBOT_SYSTEM_H_
#define _ROBOT_SYSTEM_H_

struct robot_system
{
	u08 flags;
	struct rs_polar virtual_pwm;
	struct rs_polar virtual_encoders;
	f64 ratio_mot_ext;

	/* Motor encoders */
        s32 (*left_mot_encoder)(void *);
        void* left_mot_encoder_param;
	f64 left_ext_gain;

        s32 (*right_mot_encoder)(void *);
        void* right_mot_encoder_param;
	f64 right_ext_gain;

	/* External encoders */
        s32 (*left_ext_encoder)(void *);
        void* left_ext_encoder_param;
	f64 left_mot_gain;

        s32 (*right_ext_encoder)(void *);
        void* right_ext_encoder_param;
	f64 right_mot_gain;
	
	/* PWM */
	void (*left_pwm)(void *, s32);
	void *left_pwm_param;
	void (*right_pwm)(void *, s32);
	void *right_pwm_param;
};

/** Set the structure to 0 */
void rs_init( struct robot_system * );

/**** ACCESSORS */

/** define ratio between mot and ext track. (track_mot / track_ext) */
void rs_set_ratio(struct robot_system * rs, double ratio);

/** define left PWM function and param */
void rs_set_left_pwm(struct robot_system * rs, void (*left_pwm)(void *, s32), void *left_pwm_param);

/** define right PWM function and param */
void rs_set_right_pwm(struct robot_system * rs, void (*right_pwm)(void *, s32), void *right_pwm_param);

/** define left motor encoder function and param */
void rs_set_left_mot_encoder(struct robot_system * rs, s32 (*left_mot_encoder)(void *), 
			     void *left_mot_encoder_param, double gain);

/** define right motor encoder function and param */
void rs_set_right_mot_encoder(struct robot_system * rs, s32 (*right_mot_encoder)(void *), 
			      void *right_mot_encoder_param, double gain);

/** define left external encoder function and param */
void rs_set_left_ext_encoder(struct robot_system * rs, s32 (*left_ext_encoder)(void *), 
			     void *left_ext_encoder_param, double gain);

/** define right external encoder function and param */
void rs_set_right_ext_encoder(struct robot_system * rs, s32 (*right_ext_encoder)(void *), 
			      void *right_ext_encoder_param, double gain);

/**** Virtual encoders and PWM */

/** 
 * set the real pwms according to the specified angle (it also
 * depends on the last distance command sent) 
 */
void rs_set_angle(void * rs, s32 angle);

/** 
 * set the real pwms according to the specified distance (it also
 * depends on the last angle command sent) 
 */
void rs_set_distance(void * rs, s32 distance);

/** 
 * get the virtual angle according to real encoders value. 
 */
s32 rs_get_angle(void * rs);

/** 
 * get the virtual distance according to real encoders value. 
 */
s32 rs_get_distance(void * rs);


/** 
 * Read the encoders, and update internal virtual counters. Call this
 * function is needed before reading the virtual encoders. The program
 * will decide if it the external encoders or the motor encoders are
 * taken in account (depending on flags, but not yet)
 */
void rs_update(void * rs);


#endif /* #ifndef _ROBOT_SYSTEM */
